def __is_defined_in_submodule(submodule_name_prefix, x):
    """
    Utility method to return True if x is not a module and its module name starts with submodule_name_prefix
    e.g. 'autoclass.autohash_'.

    :param submodule_name_prefix:
    :param x:
    :return:
    """
    # if _inspect.ismodule(x):
    #     return False
    # else:
    from inspect import getmodule
    m = getmodule(x)
    if m is None:
        return True
    elif hasattr(m, '__name__') and m.__name__.startswith(submodule_name_prefix):
        return True
    else:
        return False


def __get_submodule_symbols(submodule_full_name, submodule):

    from inspect import getmembers, ismodule
    all_ = []

    for x_name, symbol in getmembers(submodule):
        if not x_name.startswith('_'):
            if __is_defined_in_submodule(submodule_full_name, symbol):
                # print('{} is exported'.format(x_name))
                if ismodule(symbol):
                    all_ = all_ + __get_submodule_symbols(submodule_full_name + '.' + x_name, symbol)
                else:
                    all_.append(x_name)

    return all_


def __get_all_submodules_symbols(pkg_name, submodules_to_export):
    """
    Generates the list of symbol names that can be used in the `__all__` variable in init.py
    The list is created from a list of submodules.

    All symbols in these submodules that are not private and that are actually defined in there, get in the list.
    The submodules themselves end up in the list.

    Note that this function should only be used if you also actually import those symbols in the init.py, so that they
    are actually visible at package root level.

    :param submodules_to_export: a list of submodule names to export
    :return:
    """
    from copy import copy

    from importlib import import_module

    # first create a copy of the submodules list
    all_ = copy(submodules_to_export)

    # then for each submodule add the symbols that are declared in this submodule
    for submodule in submodules_to_export:
        submodule_full_name = pkg_name + '.' + submodule
        imported_module = import_module(submodule_full_name)
        # print(imported_module.__name__)
        all_ = all_ + __get_submodule_symbols(submodule_full_name, imported_module)

    return all_


def __remove_all_external_symbols(pkg_name, globs):
    for x_name in list(globs.keys()):
        if not __is_defined_in_submodule(pkg_name, globs[x_name]):
            del globs[x_name]



import pytest
from autoclass import autoargs, autoprops, setter_override
from valid8 import Boolean, ValidationError


def test_readme_index_basic():
    """ First basic example in the doc """

    from autoclass import autoclass

    @autoclass
    class House:
        def __init__(self, name, nb_floors=1):
            pass

    a = House('my_house', 3)
    assert str(a) == "House({'name': 'my_house', 'nb_floors': 3})"
    assert [att for att in a.keys()] == ['name', 'nb_floors']
    assert {a, a} == {a}
    assert a == {'name': 'my_house', 'nb_floors': 3}


def test_readme_index_basic2():
    """ Second basic example in the doc: adding setter override """

    from autoclass import autoclass, setter_override

    global t
    t = ''

    @autoclass
    class House:
        def __init__(self, name, nb_floors=1):
            pass

        @setter_override
        def nb_floors(self, nb_floors=1):
            global t
            t = 'Set nb_floors to {}'.format(nb_floors)
            self._nb_floors = nb_floors

    assert t == ''

    obj = House('my_house')
    assert t == 'Set nb_floors to 1'

    obj.nb_floors = 3
    assert t == 'Set nb_floors to 3'


def test_readme_enforce_simple():
    """ Example in the doc with enforce """

    from autoclass import autoclass

    # we use enforce runtime checker for this example
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    @runtime_validation
    @autoclass
    class House:
        def __init__(self, name: str, nb_floors: int = 1):
            pass

    obj = House('my_house')

    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError) as exc_info:
        obj.nb_floors = 'red'
    assert exc_info.value.args[0] == "\n  The following runtime type errors were encountered:\n" \
                                     "       Argument 'nb_floors' was not of type <class 'int'>. " \
                                     "Actual type was str."


def test_readme_index_pytypes_simple():
    """ Example in the doc with pytypes """

    from autoclass import autoclass
    from pytypes import typechecked

    @typechecked
    @autoclass
    class House:
        # the constructor below is decorated with PEP484 type hints
        def __init__(self, name: str, nb_floors: int = 1):
            pass

    obj = House('my_house')

    from pytypes import InputTypeError
    with pytest.raises(InputTypeError) as exc_info:
        obj.nb_floors = 'red'
    assert exc_info.value.args[0] == "\n  autoclass.tests.test_readme_index." \
                                     "test_readme_index_pytypes_simple.<locals>.House.nb_floors/setter" \
                                     "\n  called with incompatible types:\nExpected: Tuple[int]\nReceived: Tuple[str]"


def test_readme_index_valid8_simple():
    """ Example in the doc with valid8 """

    from autoclass import autoclass
    from mini_lambda import s, x, l, Len
    from valid8 import validate_arg, instance_of, is_multiple_of, InputValidationError

    # Here we define our 2 validation errors
    class InvalidNameError(InputValidationError):
        help_msg = 'name should be a non-empty string'

    class InvalidSurfaceError(InputValidationError):
        help_msg = 'Surface should be between 0 and 10000 and be a multiple of 100.'

    @autoclass
    class House:

        @validate_arg('name', instance_of(str), Len(s) > 0,
                      error_type=InvalidNameError)
        @validate_arg('surface', (x >= 0) & (x < 10000), is_multiple_of(100),
                      error_type=InvalidSurfaceError)
        def __init__(self, name, surface=None):
            pass

    obj = House('sweet home', 200)

    obj.surface = None  # Valid (surface is nonable by signature)
    with pytest.raises(InvalidNameError):
        obj.name = 12  # InvalidNameError
    with pytest.raises(InvalidSurfaceError):
        obj.surface = 10000  # InvalidSurfaceError


def test_readme_index_enforce_valid8():
    """ Makes sure that the code in the documentation page is correct for the enforce + valid8 example """

    from autoclass import autoclass

    # Imports - for type validation
    from numbers import Integral
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # type validation will accept subclasses too

    # Imports - for value validation
    from mini_lambda import s, x, Len
    from valid8 import validate_arg, is_multiple_of, InputValidationError

    # 2 custom validation errors for valid8
    class InvalidName(InputValidationError):
        help_msg = 'name should be a non-empty string'

    class InvalidSurface(InputValidationError):
        help_msg = 'Surface should be between 0 and 10000 and be a multiple of 100.'

    @runtime_validation
    @autoclass
    class House:
        @validate_arg('name', Len(s) > 0,
                      error_type=InvalidName)
        @validate_arg('surface', (x >= 0) & (x < 10000), is_multiple_of(100),
                      error_type=InvalidSurface)
        def __init__(self, name: str, surface: Integral = None):
            pass

    obj = House('sweet home', 200)

    obj.surface = None  # Valid (surface is nonable by signature)
    with pytest.raises(InvalidName):
        obj.name = ''  # InvalidNameError
    with pytest.raises(InvalidSurface):
        obj.surface = 10000  # InvalidSurfaceError


def test_readme_pycontracts_simple():
    """ Simple test with pycontracts """

    from autoclass import autoclass
    from contracts import contract, ContractNotRespected

    @autoclass
    class House:

        @contract(name='str[>0]',
                  surface='None|(int,>=0,<10000)')
        def __init__(self, name, surface):
            pass

    obj = House('sweet home', 200)

    obj.surface = None  # Valid (surface is nonable by signature)
    with pytest.raises(ContractNotRespected):
        obj.name = ''  # InvalidNameError
    with pytest.raises(ContractNotRespected):
        obj.surface = 10000  # InvalidSurfaceError


def test_readme_old_way():
    """ Makes sure that the code in the documentation page is correct for the 'old way' of writing classes """

    from autoclass import check_var
    from numbers import Real, Integral
    from typing import Optional, Union
    from valid8 import Boolean

    class HouseConfiguration(object):
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            self.name = name
            self.surface = surface
            self.nb_floors = nb_floors
            self.with_windows = with_windows

        # --name
        @property
        def name(self):
            return self._name

        @name.setter
        def name(self, name: str):
            check_var(name, var_name='name', var_types=str)
            self._name = name

        # --surface
        @property
        def surface(self) -> Real:
            return self._surface

        @surface.setter
        def surface(self, surface: Real):
            check_var(surface, var_name='surface', var_types=Real, min_value=0, min_strict=True)
            self._surface = surface

        # --nb_floors
        @property
        def nb_floors(self) -> Optional[Integral]:
            return self._nb_floors

        @nb_floors.setter
        def nb_floors(self, nb_floors: Optional[Integral]):
            check_var(nb_floors, var_name='nb_floors', var_types=Integral, enforce_not_none=False)
            self._surface = nb_floors  # !**

        # --with_windows
        @property
        def with_windows(self) -> Boolean:
            return self._with_windows

        @with_windows.setter
        def with_windows(self, with_windows: Boolean):
            check_var(with_windows, var_name='with_windows', var_types=Boolean)
            self._with_windows = with_windows

    HouseConfiguration('test', 0.1)


def test_readme_pycontracts_complex():
    """ Makes sure that the code in the documentation page is correct for the PyContracts example """

    from contracts import contract, ContractNotRespected
    from numbers import Real, Integral
    from typing import Optional

    @autoprops
    class HouseConfiguration(object):
        @autoargs
        @contract(name='str[>0]',
                  surface='(int|float),>=0',
                  nb_floors='None|int',
                  with_windows='bool')
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface - no need to repeat the @contract
        @setter_override
        def surface(self, surface: Real):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 0.1)
    t.nb_floors = None
    with pytest.raises(ContractNotRespected):
        t.nb_floors = 2.2
    with pytest.raises(ContractNotRespected):
        t.surface = -1


@pytest.mark.skip(reason="open bug in pytypes https://github.com/Stewori/pytypes/issues/19")
def test_readme_pytypes_validate_complex():
    """ A more complex pytypes + valid8 example """

    from autoclass import autoclass, setter_override
    from valid8 import Boolean, validate_io, minlens, gt
    from numbers import Real, Integral
    from typing import Optional

    # we use pytypes for this example
    from pytypes import typechecked

    @typechecked
    @autoclass
    class HouseConfiguration(object):

        @validate_io(name=minlens(0),
                     surface=gt(0))
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation or other things
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from pytypes import InputTypeError
    with pytest.raises(InputTypeError):
        t.nb_floors = 2.2

    # Value validation works
    with pytest.raises(ValidationError):
        t.surface = -1

    # Value validation works in constructor
    with pytest.raises(ValidationError):
        HouseConfiguration('', 12, 2)

    # str and repr work
    assert str(t) == repr(t)
    # "HouseConfiguration({'nb_floors': None, 'with_windows': False, 'surface': 12, 'name': 'test'})"
    # assert eval(repr(t)) == t does not work !

    # dict work
    assert t == {'name': 'test', 'nb_floors': None, 'surface': 12, 'with_windows': False}
    assert t == dict(name='test', nb_floors=None, surface=12, with_windows=False)
    t.keys()
    for k, v in t.items():
        print(str(k) + ': ' + str(v))

    # TODO this is an open bug in pytypes https://github.com/Stewori/pytypes/issues/19
    HouseConfiguration.from_dict({'name': 'test2', 'surface': 1})



from inspect import signature
from collections import Sequence
from typing import Tuple, Callable, Union

from autoclass.utils_decoration import _create_function_decorator__robust_to_args
from autoclass.utils_include_exclude import _sieve
from autoclass.var_checker import check_var
from decorator import decorate


def autoargs(include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None):
    """
    Defines a decorator with parameters, to automatically assign the inputs of a function to self PRIOR to executing 
    the function. In other words:
    
    ```
    @autoargs
    def myfunc(a):
        print('hello')
    ```
    
    will create the equivalent of
    
    ```
    def myfunc(a):
        self.a = a
        print('hello')
    ```
    
    Initial code from http://stackoverflow.com/questions/3652851/what-is-the-best-way-to-do-automatic-attribute-assignment-in-python-and-is-it-a#answer-3653049

    :param include: a tuple of attribute names to include in the auto-assignment. If None, all arguments will be
    included by default
    :param exclude: a tuple of attribute names to exclude from the auto-assignment. In such case, include should be None
    :return:
    """
    return _create_function_decorator__robust_to_args(autoargs_decorate, include, exclude=exclude)


def autoargs_decorate(func: Callable, include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None) \
        -> Callable:
    """
    Defines a decorator with parameters, to automatically assign the inputs of a function to self PRIOR to executing 
    the function. This is the inline way to apply the decorator 
    
    ```
    myfunc2 = autoargs_decorate(myfunc)
    ```
    
    
    See autoargs for details.

    :param func: the function to wrap
    :param include: a tuple of attribute names to include in the auto-assignment. If None, all arguments will be
    included by default
    :param exclude: a tuple of attribute names to exclude from the auto-assignment. In such case, include should be None
    :return:
    """

    # (0) first check parameters
    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')
    check_var(include, var_name='include', var_types=[str, Sequence], enforce_not_none=False)
    check_var(exclude, var_name='exclude', var_types=[str, Sequence], enforce_not_none=False)

    # (1) then retrieve function signature
    # attrs, varargs, varkw, defaults = getargspec(func)
    func_sig = signature(func)

    # check that include/exclude dont contain names that are incorrect
    if include is not None:
        incorrect = set(list([include] if isinstance(include, str) else include)) - set(func_sig.parameters.keys())
        if len(incorrect) > 0:
            raise ValueError('@autoargs definition exception: include contains \'' + str(incorrect) + '\' that is/are '
                            'not part of signature for ' + str(func))
    if exclude is not None:
        incorrect = set([exclude] if isinstance(exclude, str) else exclude) - set(func_sig.parameters.keys())
        if len(incorrect) > 0:
            raise ValueError('@autoargs definition exception: exclude contains \'' + str(incorrect) + '\' that is/are '
                             'not part of signature for ' + str(func))

    # TODO this should be in @autoslots decorator at class level, not here.
    # (2) Optionally lock the class only for the provided fields
    # Does not work for the moment. Besides locking fields seems to have issues with pickle serialization
    # so we'd rather not propose this option.
    #
    # See 'attrs' project for this kind of advanced features https://github.com/python-attrs/attrs
    #
    # if lock_class_fields:
    #     if signature_varkw:
    #         raise Exception('cant lock field names with variable kwargs')
    #     else:
    #         object_type = get_class_that_defined_method(func)
    #         if include:
    #             fields = include
    #         else:
    #             fields = signature_attrs[1:]
    #             if signature_varargs:
    #                 fields.append(signature_varargs)
    #             if exclude:
    #                 for a in exclude:
    #                     fields.remove(a)
    #
    #         # right now, doesnot work
    #         _lock_fieldnames_class(object_type, field_names=tuple(fields))

    # (3) Finally, create a wrapper around the function so that all attributes included/not excluded are
    # set to self BEFORE executing the function.
    # -- NOTE: we used @functools.wraps(func), we now use 'decorate' to have a wrapper that has the same signature
    def init_wrapper(init_func, self, *args, **kwargs):

        # match the received arguments with the signature to know who is who, and add default values to get a full list
        bound_values = func_sig.bind(self, *args, **kwargs)
        bound_values.apply_defaults()

        # Assign to self the ones that needs to
        for att_name, att_value in bound_values.arguments.items():
            if _sieve(att_name, include=include, exclude=exclude):
                # value = a normal value, or cur_kwargs as a whole
                setattr(self, att_name, att_value)
            # The behaviour below is removed as it is too complex to explain
            # else:
            #     if func_sig.parameters[att_name].kind == Parameter.VAR_KEYWORD:
            #         # if the attribute is variable-length keyword argument we can try to find a matching key inside it
            #         # each item is handled independently (if func signature contains the kw args names such as a, b)
            #         for name, value in att_value.items():
            #             if _sieve(name, include=include, exclude=exclude):
            #                 setattr(self, name, value)

        # finally execute the constructor function
        return init_func(self, *args, **kwargs)

    # return wrapper
    return decorate(func, init_wrapper)



from autoclass.utils_init import __remove_all_external_symbols, __get_all_submodules_symbols

__PACKAGE_NAME = 'autoclass'
__SUBMODULES_TO_EXPORT = ['autoclass_', 'autoargs_', 'autoprops_', 'autodict_', 'autohash_', 'var_checker']
# TODO we could rather rely on a regexp mechanism

# (1) allow users to do
#     import <package> as p and then p.<symbol>
__all__ = __get_all_submodules_symbols(__PACKAGE_NAME, __SUBMODULES_TO_EXPORT)
# Note: this is one way to do it, but it would be simpler to check the names in globals() at the end of this file.

# (2) allow users to do
#     from <package> import <symbol>
#
# The following works, but unfortunately IDE like pycharm do not understand
from autoclass.autoargs_ import *
from autoclass.autoclass_ import *
from autoclass.autodict_ import *
from autoclass.autohash_ import *
from autoclass.autoprops_ import *
from autoclass.var_checker import *

# remove all symbols that were imported above but do not belong in this package
__remove_all_external_symbols(__PACKAGE_NAME, globals())

# Otherwise exhaustive list would be required, which is sad
# from autoclass.autoargs_ import autoargs, autoargs_decorate
# from autoclass.autoclass_ import autoclass, autoclass_decorate
# from autoclass.autodict_ import autodict, autodict_override, autodict_decorate, autodict_override_decorate, \
#     print_ordered_dict
# from autoclass.autohash_ import autohash, autohash_decorate
# from autoclass.autoprops_ import IllegalGetterSignatureException, IllegalSetterSignatureException, autoprops, \
#     autoprops_decorate, DuplicateOverrideError, getter_override, setter_override, autoprops_override_decorate
# from autoclass.var_checker import check_var, MissingMandatoryParameterException

# print(__all__)
# print(globals().keys())
# print('Done')



from inspect import isclass
from typing import Callable


class AutoclassDecorationException(Exception):
    pass


def _check_known_decorators(typ, calling_decorator: str) -> bool:
    """
    Checks that a given type is not already decorated by known decorators that may cause trouble.
    If so, it raises an Exception
    :return:
    """
    for member in typ.__dict__.values():
        if hasattr(member, '__enforcer__'):
            raise AutoclassDecorationException('It seems that @runtime_validation decorator was applied to type <'
                                               + str(typ) + '> BEFORE ' + calling_decorator + '. This is not supported '
                                               'as it may lead to counter-intuitive behaviour, please change the order '
                                               'of the decorators on <' + str(typ) + '>')


def _call_decorator_with_or_without_args(decorator_function_impl: Callable, objectIsFunction: bool, *args, **kwargs):
    """
    Utility function for all decorators: handles the fact that 
    * if the decorator has been called with arguments, then we have to return a decorator factory
    * if the decorator has been called without arguments, then we have to directly return the decoration result

    :param decorator_function_impl:
    :param args:
    :param kwargs:
    :return:
    """

    if len(args) > 0 and ((objectIsFunction and callable(args[0])) or (not objectIsFunction and isclass(args[0]))):
        # we have been called without argument. In that case 'args[0]' contains the object_to_decorate
        return decorator_function_impl(*args, **kwargs)
    else:
        # called with arguments : lets return a factory function
        def f(object_to_decorate):
            return decorator_function_impl(object_to_decorate, *args, **kwargs)
        return f


def _create_function_decorator__robust_to_args(decorator_impl: Callable, *args, **kwargs):
    """
    Utility method to implement a decorator for functions based on the provided implementation and arguments
    
    :param decorator_impl: a function taking as first argument a function
    :param args: 
    :param kwargs: 
    :return: 
    """
    return _call_decorator_with_or_without_args(decorator_impl, True, *args, **kwargs)


def _create_class_decorator__robust_to_args(decorator_impl: Callable, *args, **kwargs):
    """
    Utility method to create a decorator for classes based on the provided implementation and arguments
    
    :param decorator_impl: 
    :param args: 
    :param kwargs: 
    :return: 
    """
    return _call_decorator_with_or_without_args(decorator_impl, False, *args, **kwargs)


# def apply_on_func_args(func, cur_args, cur_kwargs,
#                        signature_attrs, signature_defaults, signature_varargs, signature_varkw,
#                        func_to_apply, func_to_apply_paramers_dict):
#     """
#     Applies func_to_apply on each argument of func according to what's received in current call (cur_args, cur_kwargs).
#     For each argument of func named 'att' in its signature, the following method is called:
#
#     `func_to_apply(cur_att_value, func_to_apply_paramers_dict[att], func, att_name)`
#
#     :param func:
#     :param cur_args:
#     :param cur_kwargs:
#     :param signature_attrs:
#     :param signature_defaults:
#     :param signature_varargs:
#     :param signature_varkw:
#     :param func_to_apply:
#     :param func_to_apply_paramers_dict:
#     :return:
#     """
#     # handle default values and kw arguments
#     for attr, val in zip(reversed(signature_attrs), reversed(signature_defaults or [])):
#         if attr in func_to_apply_paramers_dict.keys():
#             # set default or provided value
#             if attr in cur_kwargs.keys():
#                 # provided: we never seem to enter here, why ? maybe depends on the version of python
#                 func_to_apply(cur_kwargs[attr], func_to_apply_paramers_dict[attr], func, attr)
#             else:
#                 # default
#                 func_to_apply(val, func_to_apply_paramers_dict[attr], func, attr)
#
#     # handle positional arguments
#     for attr, val in zip(signature_attrs, cur_args):
#         if attr in func_to_apply_paramers_dict.keys():
#             func_to_apply(val, func_to_apply_paramers_dict[attr], func, attr)
#
#     # handle varargs : since we dont know their name, they can only be validated as a whole
#     if signature_varargs:
#         remaining_args = cur_args[len(signature_attrs):]
#         if signature_varargs in func_to_apply_paramers_dict.keys():
#             func_to_apply(remaining_args, func_to_apply_paramers_dict[signature_varargs], func, signature_varargs)
#
#     # handle varkw : since we know their names, they can be validated either as a whole or independently
#     if signature_varkw:
#         # either the item is the whole dictionary (if func signature is generic such as in **kwargs)
#         if signature_varkw in func_to_apply_paramers_dict.keys():
#             # value = the cur_kwargs as a whole
#             func_to_apply(cur_kwargs, func_to_apply_paramers_dict[signature_varkw], func, signature_varkw)
#         else:
#             # or each item is handled independently (if func signature contains the kw args names such as a, b)
#             for attr, val in cur_kwargs.items():
#                 if attr in func_to_apply_paramers_dict.keys():
#                     func_to_apply(val, func_to_apply_paramers_dict[attr], func, attr)


# def apply_on_each_func_args_sig(func, cur_args, cur_kwargs, sig: Signature,
#                                 func_to_apply, func_to_apply_paramers_dict):
#     """
#     Applies func_to_apply on each argument of func according to what's received in current call (cur_args, cur_kwargs).
#     For each argument of func named 'att' in its signature, the following method is called:
#
#     `func_to_apply(cur_att_value, func_to_apply_paramers_dict[att], func, att_name)`
#
#     :param func:
#     :param cur_args:
#     :param cur_kwargs:
#     :param sig:
#     :param func_to_apply:
#     :param func_to_apply_paramers_dict:
#     :return:
#     """
#
#     # match the received arguments with the signature to know who is who
#     bound_values = sig.bind(*cur_args, **cur_kwargs)
#
#     # add the default values in here to get a full list
#     bound_values.apply_defaults()
#
#     for att_name, att_value in bound_values.arguments.items():
#         if att_name in func_to_apply_paramers_dict.keys():
#             # value = a normal value, or cur_kwargs as a whole
#             func_to_apply(att_value, func_to_apply_paramers_dict[att_name], func, att_name)
#         # The behaviour below is removed as it is too complex to explain
#         # else:
#         #    if sig.parameters[att_name].kind == Parameter.VAR_KEYWORD:
#                 # if the attribute is variable-length keyword argument we can try to find a matching key inside it
#                 # each item is handled independently (if func signature contains the kw args names such as a, b)
#         #        for name, value in att_value.items():
#         #            if name in func_to_apply_paramers_dict.keys():
#         #                func_to_apply(value, func_to_apply_paramers_dict[name], func, name)



from collections import Sequence
from inspect import signature
from typing import Tuple, Union, TypeVar  # do not fail on import Type for compatibility with earlier python 3.5

from autoclass.var_checker import check_var
from autoclass.utils_include_exclude import _sieve
from autoclass.utils_reflexion import get_constructor
from autoclass.utils_decoration import _create_class_decorator__robust_to_args, _check_known_decorators


def autohash(include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None,
             only_constructor_args: bool = False, only_public_fields: bool = False):
    """
    A decorator to makes objects of the class implement __hash__, so that they can be used correctly for example in
    sets.
    
    Parameters allow to customize the list of attributes that are taken into account in the hash.

    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if False (default), all fields will be included in the hash, even if they are defined
    in the constructor or dynamically. If True, only constructor arguments will be included in the hash, not any other
    field that would be created in the constructor or dynamically. Please note that this behaviour is the opposite from
    @autodict.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False (default), all fields are used in the hash. Otherwise, class-private fields will
    not be taken into account in the hash. Please note that this behaviour is the opposite from @autodict.
    :return:
    """
    return _create_class_decorator__robust_to_args(autohash_decorate, include, exclude=exclude,
                                                   only_constructor_args=only_constructor_args,
                                                   only_public_fields=only_public_fields)


T = TypeVar('T')


def autohash_decorate(cls: 'Type[T]', include: Union[str, Tuple[str]] = None, exclude: Union[str, Tuple[str]] = None,
                      only_constructor_args: bool = False, only_public_fields: bool = False) -> 'Type[T]':
    """
    To automatically generate the appropriate methods so that objects of this class are hashable,
    manually, without using @autohash decorator.

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if False (default), all fields will be included in the hash, even if they are defined
    in the constructor or dynamically. If True, only constructor arguments will be included in the hash, not any other
    field that would be created in the constructor or dynamically. Please note that this behaviour is the opposite from
    @autodict.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False (default), all fields are used in the hash. Otherwise, class-private fields will
    not be taken into account in the hash. Please note that this behaviour is the opposite from @autodict.
    :return:
    """

    # first check that we do not conflict with other known decorators
    _check_known_decorators(cls, '@autohash')

    # perform the class mod
    _execute_autohash_on_class(cls, include=include, exclude=exclude, only_constructor_args=only_constructor_args,
                               only_public_fields=only_public_fields)

    return cls


def _execute_autohash_on_class(object_type: 'Type[T]', include: Union[str, Tuple[str]]=None,
                               exclude: Union[str, Tuple[str]]=None, only_constructor_args: bool = False,
                               only_public_fields: bool = False):
    """
    A decorator to make objects of the class implement __hash__, so that they can be used correctly for example in
    sets.

    Parameters allow to customize the list of attributes that are taken into account in the hash.

    :param object_type: the class on which to execute.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if False (default), all fields will be included in the hash, even if they are defined
    in the constructor or dynamically. If True, only constructor arguments will be included in the hash, not any other
    field that would be created in the constructor or dynamically. Please note that this behaviour is the opposite from
    @autodict.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False (default), all fields are used in the hash. Otherwise, class-private fields will
    not be taken into account in the hash. Please note that this behaviour is the opposite from @autodict.
    :return:
    """

    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')
    check_var(include, var_name='include', var_types=[str, Sequence], enforce_not_none=False)
    check_var(exclude, var_name='exclude', var_types=[str, Sequence], enforce_not_none=False)

    # Override hash method if not already implemented
    if not hasattr(object_type, '__hash__') or object_type.__hash__ is None or object_type.__hash__ == object.__hash__:

        if only_constructor_args:
            # a. Find the __init__ constructor signature
            constructor = get_constructor(object_type, allow_inheritance=True)
            s = signature(constructor)

            # b. Collect all attributes that are not 'self' and are included and not excluded
            added = []
            # we assume that the order of attributes will always be the same here....
            for attr_name in s.parameters.keys():
                if _sieve(attr_name, include=include, exclude=exclude):
                    added.append(attr_name)

            # c. Finally build the method
            def __hash__(self):
                """
                Generated by @autohash.
                Implements the __hash__ method by hashing a tuple of selected attributes
                :param self:
                :return:
                """
                # note: we prepend a unique hash for the class  > NO, it is more intuitive to not do that.
                # return hash(tuple([type(self)] + [getattr(self, att_name) for att_name in added]))
                return hash(tuple(getattr(self, att_name) for att_name in added))

        else:
            # ** all dynamic fields are allowed
            if include is None and exclude is None and not only_public_fields:

                # easy: all of vars values is included in the hash
                def __hash__(self):
                    """
                    Generated by @autohash.
                    Implements the __hash__ method by hashing vars(self).values()
                    :param self:
                    :return:
                    """
                    # note: we prepend a unique hash for the class  > NO, it is more intuitive to not do that.
                    # return hash(tuple([type(self)] + list(vars(self).values())))
                    return hash(tuple(vars(self).values()))

            else:
                # harder: dynamic filters
                # private_name_prefix = '_' + object_type.__name__ + '_'
                private_name_prefix = '_'

                def __hash__(self):
                    """
                    Generated by @autohash.
                    Implements the __hash__ method by hashing the tuple of included/not excluded field values, possibly
                    not including the private fields if `only_public_fields` was set to True

                    :param self:
                    :return:
                    """
                    # note: we prepend a unique hash for the class > NO, it is more intuitive to not do that.
                    # to_hash = [type(self)]
                    to_hash = []

                    for att_name, att_value in vars(self).items():
                        if _sieve(att_name, include=include, exclude=exclude):
                            if not only_public_fields \
                                    or (only_public_fields and not att_name.startswith(private_name_prefix)):
                                to_hash.append(att_value)

                    return hash(tuple(to_hash))

        # now set the method on the class
        object_type.__hash__ = __hash__

    return
